import os
import re
import unicodedata
from contextlib import contextmanager
from datetime import datetime, timedelta, timezone
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
)  # noqa: F401

import pytz

T = TypeVar("T")

UUID_REGEX_PATTERN = (
    r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"
)
EMAIL_REGEX_PATTERN = (
    r"^([-!#$%&'*+/=?^_`{}|~0-9a-zA-Z]+"
    r"(\.[-!#$%&'*+/=?^_`{}|~0-9a-zA-Z]+)*|^\""
    r"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016-\177])*\")"
    r"@(?:[a-zA-Z0-9\[\]](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)+[a-zA-Z0-9\[\]]{2,6}\.?$"
)


def find(iterable: Iterable[T], predicate: Callable[[T], bool]) -> Optional[T]:
    """Return the first item in the iterable that matches the predicate function."""
    for i in iterable:
        if predicate(i):
            return i

    return None


def is_uuid(value: str) -> bool:
    """
    Check if the string value is a proper UUID string.

    UUID check is performed based on a regex pattern:
        `r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$"`

    :return: True if there is a match, otherwise False
    :rtype: bool
    """
    return bool(re.match(UUID_REGEX_PATTERN, str(value)))


def is_url(value: str) -> bool:
    """
    Return whether or not given value is a valid URL.

    If the value is valid URL this function returns ``True``.

    This validator implementation is taken and augmented from https://github.com/kvesteri/validators/
    Copyright (c) 2013-2014 Konsta Vesterinen, It is Licenced with the MIT license.

    This validator is based on the wonderful `URL validator of dperini`_.

    .. _URL validator of dperini:
        https://gist.github.com/dperini/729294

    :param value: an alleged url
    :type value: Text
    :return: True if it is an URL, otherwise False
    :rtype: bool

    Examples
    --------
    >>> is_url('http://foobar.dk')
    True
    >>> is_url('ftp://foobar.dk')
    True
    >>> is_url('http://10.0.0.1')
    True
    >>> is_url('http://foobar.d')
    False

    """
    ip_middle_octet = r"(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5]))"
    ip_last_octet = r"(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))"

    regex = re.compile(  # noqa: W605
        r"^"
        # protocol identifier
        r"(?:(?:https?|ftp)://)"
        # user:pass authentication
        r"(?:[-a-z\u00a1-\uffff0-9._~%!$&'()*+,;=:]+"
        r"(?::[-a-z0-9._~%!$&'()*+,;=:]*)?@)?"
        r"(?:"
        r"(?P<private_ip>"
        # IP address exclusion
        # private & local networks
        r"(?:(?:10|127)" + ip_middle_octet + r"{2}" + ip_last_octet + r")|"
        r"(?:(?:169\.254|192\.168)" + ip_middle_octet + ip_last_octet + r")|"
        r"(?:172\.(?:1[6-9]|2\d|3[0-1])" + ip_middle_octet + ip_last_octet + r"))"
        r"|"
        # private & local hosts
        r"(?P<private_host>" r"(?:localhost))" r"|"
        # IP address dotted notation octets
        # excludes loopback network 0.0.0.0
        # excludes reserved space >= 224.0.0.0
        # excludes network & broadcast addresses
        # (first & last IP address of each class)
        r"(?P<public_ip>"
        r"(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])"
        r"" + ip_middle_octet + r"{2}"
        r"" + ip_last_octet + r")"
        r"|"
        # IPv6 RegEx from https://stackoverflow.com/a/17871737
        r"\[("
        # 1:2:3:4:5:6:7:8
        r"([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|"
        # 1::                              1:2:3:4:5:6:7::
        r"([0-9a-fA-F]{1,4}:){1,7}:|"
        # 1::8             1:2:3:4:5:6::8  1:2:3:4:5:6::8
        r"([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|"
        # 1::7:8           1:2:3:4:5::7:8  1:2:3:4:5::8
        r"([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|"
        # 1::6:7:8         1:2:3:4::6:7:8  1:2:3:4::8
        r"([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|"
        # 1::5:6:7:8       1:2:3::5:6:7:8  1:2:3::8
        r"([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|"
        # 1::4:5:6:7:8     1:2::4:5:6:7:8  1:2::8
        r"([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|"
        # 1::3:4:5:6:7:8   1::3:4:5:6:7:8  1::8
        r"[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|"
        # ::2:3:4:5:6:7:8  ::2:3:4:5:6:7:8 ::8       ::
        r":((:[0-9a-fA-F]{1,4}){1,7}|:)|"
        # fe80::7:8%eth0   fe80::7:8%1
        # (link-local IPv6 addresses with zone index)
        r"fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|"
        r"::(ffff(:0{1,4}){0,1}:){0,1}"
        r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
        # ::255.255.255.255   ::ffff:255.255.255.255  ::ffff:0:255.255.255.255
        # (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
        r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|"
        r"([0-9a-fA-F]{1,4}:){1,4}:"
        r"((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}"
        # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33
        # (IPv4-Embedded IPv6 Address)
        r"(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])" r")\]|"
        # host name
        r"(?:(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)"
        # domain name
        r"(?:\.(?:[a-z\u00a1-\uffff0-9]-?)*[a-z\u00a1-\uffff0-9]+)*"
        # TLD identifier
        r"(?:\.(?:[a-z\u00a1-\uffff]{2,}))" r")"
        # port number
        r"(?::\d{2,5})?"
        # resource path
        "(?:/[-a-z\u00a1-\uffff0-9._~%!$&'()*+,;=:@/]*)?"
        # query string
        r"(?:\?\S*)?"
        # fragment
        r"(?:#\S*)?" r"$",
        re.UNICODE | re.IGNORECASE,
    )
    return bool(regex.match(value))


def is_valid_email(value: str) -> bool:
    """
    Return whether or not given value is a valid email address.

    If the value is valid email this function returns ``True``.
    The email will only be checked if it contains the right form, not if the email
    address exist. There are many and better ways to check if the email is valid, but
    in this case we check if the email has the form of `name@domain.tld`.

    :param value: an alleged email
    :type value: Text
    :return: True if it is an email, otherwise False
    :rtype: bool

    Examples
    --------
    >>> is_valid_email("valid_email-address@example.com")
    True
    >>> is_valid_email("invalid@email")
    False

    """
    if re.match(EMAIL_REGEX_PATTERN, str(value)):
        return True
    else:
        return False


@contextmanager
def temp_chdir(cwd: Optional[str] = None):
    """
    Create and return a temporary directory which you can use as a context manager.

    When you are out of the context the temprorary disk gets erased.

    .. versionadded:: 2.3

    :param cwd: path to change working directory back to path when out of context
    :type cwd: basestring or None
    :return: in context a temporary directory

    Example
    -------
    >>> with temp_chdir() as temp_dir:
    >>>     # do things here
    >>>     print(temp_dir)  # etc etc
    ...
    >>> # when moving out of the context the temp_dir is destroyed
    >>> pass

    """
    from tempfile import TemporaryDirectory

    with TemporaryDirectory() as tempwd:
        origin = cwd or os.getcwd()
        os.chdir(tempwd)

        try:
            yield tempwd if os.path.exists(tempwd) else ""
        finally:
            os.chdir(origin)


def parse_datetime(value: Optional[str]) -> Optional[datetime]:
    """
    Convert datetime string to datetime object.

    Helper function to convert a datetime string found in json responses to a datetime object with timezone information.
    The server is storing all datetime strings as UTC (ZULU time). This function supports time zone offsets. When
    the input contains one, the output uses a timezone with a fixed offset from UTC.

    Inspired on the Django project. From `django.utils.dateparse.parse_datetime`. The code is copyrighted and
    licences with an MIT license in the following fashion::

        Copyright (c) Django Software Foundation and individual contributors.
        All rights reserved.

    ..versionadded 2.5:

    :param value: datetime string
    :type value: str or None
    :return: datetime of the value is well formatted. Otherwise (including if value is None) returns None
    :rtype: datetime or None
    :raises ValueError: if the value is well formatted but not a valid datetime
    """
    if value is None:
        # do not process the value
        return None

    def _get_fixed_timezone(offset):
        """Return a tzinfo instance with a fixed offset from UTC."""
        if isinstance(offset, timedelta):
            offset = offset.total_seconds() // 60
        sign = "-" if offset < 0 else "+"
        hhmm = "%02d%02d" % divmod(abs(offset), 60)
        name = sign + hhmm
        return timezone(timedelta(minutes=offset), name)

    DATETIME_RE = re.compile(
        r"(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})"
        r"[T ](?P<hour>\d{1,2}):(?P<minute>\d{1,2})"
        r"(?::(?P<second>\d{1,2})(?:\.(?P<microsecond>\d{1,6})\d{0,6})?)?"
        r"(?P<tzinfo>Z|[+-]\d{2}(?::?\d{2})?)?$"
    )

    match = DATETIME_RE.match(value)
    if match:
        kw = match.groupdict()
        if kw["microsecond"]:
            kw["microsecond"] = kw["microsecond"].ljust(6, "0")
        tzinfo = kw.pop("tzinfo")
        if tzinfo == "Z":
            tzinfo = pytz.UTC
        elif tzinfo is not None:
            offset_mins = int(tzinfo[-2:]) if len(tzinfo) > 3 else 0
            offset = 60 * int(tzinfo[1:3]) + offset_mins
            if tzinfo[0] == "-":
                offset = -offset
            tzinfo = _get_fixed_timezone(offset)
        kw = {k: int(v) for k, v in kw.items() if v is not None}
        kw["tzinfo"] = tzinfo
        return datetime(**kw)


#
# The following functions are inspired by: https://github.com/okunishinishi/python-stringcase
# License: MIT
#


def camelcase(string: str) -> str:
    """Convert string into camel case.

    Inspired by: https://github.com/okunishinishi/python-stringcase
    License: MIT

    :param string: String to convert.
    :returns: Camel case string.

    Examples
    --------
    >>> camelcase('foo_bar_baz')
    fooBarBaz
    >>> camelcase('FooBarBaz')
    fooBarBaz

    """
    string = re.sub(r"^[\-_\.]", "", str(string))
    if not string:
        return string
    return lowercase(string[0]) + re.sub(
        r"[\-_\.\s]([a-z])", lambda matched: uppercase(matched.group(1)), string[1:]
    )


def capitalcase(string: str) -> str:
    """Convert string into capital case.

    First letters will be uppercase.

    Inspired by: https://github.com/okunishinishi/python-stringcase
    License: MIT

    :param string: String to convert.
    :returns: Capital case string.

    Examples
    --------
    >>> capitalcase('foo_bar_baz')
    Foo_bar_baz
    >>> capitalcase('FooBarBaz')
    FooBarBaz

    """
    string = str(string)
    if not string:
        return string
    return uppercase(string[0]) + string[1:]


def lowercase(string: str) -> str:
    """Convert string into lower case.

    Inspired by: https://github.com/okunishinishi/python-stringcase
    License: MIT

    :param string: String to convert.
    :returns: lower case string.

    Examples
    --------
    >>> lowercase('foo_bar_baz')
    Foo_bar_baz
    >>> lowercase('FooBarBaz')
    foobarbaz

    """
    return str(string).lower()


def snakecase(string: str) -> str:
    """Convert string into snake case.

    Join punctuation with underscore

    Inspired by: https://github.com/okunishinishi/python-stringcase
    License: MIT


    :param string: String to convert.
    :returns: snake_case_string.


    Examples
    --------
    >>> snakecase('foo_bar_baz')
    foo_bar_baz
    >>> snakecase('FooBarBaz')
    foo_bar_baz

    """
    string = re.sub(r"[\-\.\s]", "_", str(string))
    if not string:
        return string
    return lowercase(string[0]) + re.sub(
        r"[A-Z]", lambda matched: "_" + lowercase(matched.group(0)), string[1:]
    )


def uppercase(string: str) -> str:
    """Convert string into upper case.

    Inspired by: https://github.com/okunishinishi/python-stringcase
    License: MIT

    :param string: String to convert.
    :returns: Upper case string.

    Examples
    --------
    >>> uppercase('foo_bar_baz')
    FOO_BAR_BAZ
    >>> uppercase('FooBarBaz')
    FOOBARBAZ

    """
    return str(string).upper()


def slugify_ref(value: str, allow_unicode: bool = False) -> str:
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.

    Remove characters that aren't alphanumerics, underscores, or hyphens. Convert to lowercase.
    Also strip leading and trailing whitespace.

    :param value: text to slugify
    :param allow_unicode: (optional) boolean to allow unicode processing, defaults to False
    :return: slugified value
    """
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
        value = re.sub(r"[^\w\s-]", "", value, flags=re.U).strip().lower()
        return re.sub(r"[-\s]+", "-", value, flags=re.U)
    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s-]", "", value).strip().lower()
    return re.sub(r"[-\s]+", "-", value)


def __dict_public__(cls: type(object)) -> Dict:
    """
    Get the __dict__ of the class `cls`, excluding 'dunder' attributes and methods.

    :param cls: class object
    :return: dictionary with public attributes

    Example
    -------
    >>> from pykechain.enums import Category
    >>> sorted(__dict_public__(cls=Category).values())
    ['INSTANCE', 'MODEL']

    """
    return {k: v for (k, v) in cls.__dict__.items() if not k.startswith("__")}


def __dict__inherited__(
    cls: type(object), stop: type(object) = type, public: Optional[bool] = True
) -> Dict:
    """
    Get all __dict__ items of the class and its superclasses up to `type`, or the `stop` class given as input.

    :param cls: class from which to retrieve the dict.
    :type cls: type(object)
    :param stop: optional class to indicate up to which superclass the inheritance should accumulate the dict.
    :type stop: type(object)
    :param public: optional flag, will only retrieve public (without double underscore) attributes and methods.
    :type public: bool
    :return: dictionary of key, value pairs
    :rtype dict

    Example
    -------
    >>> from pykechain.enums import Enum, Category
    >>> sorted(__dict__inherited__(cls=Category, stop=Enum, public=True).values())
    ['INSTANCE', 'MODEL']

    """
    if public:
        _dict = __dict_public__(cls=cls)
    else:
        _dict = cls.__dict__

    for super_class in cls.mro():
        if super_class == stop:
            break
        if public:
            super_class_dict = __dict_public__(cls=super_class)
        else:
            super_class_dict = super_class.__dict__
        _dict.update(super_class_dict)

    return _dict


def get_in_chunks(lst: Union[List, Iterable], chunk_size: int) -> Iterable:
    """
    Yield successive chunks from a list based on the chunk_size.

    :param lst: list or Iterable
    :type lst: List or Iterable
    :param chunk_size: size of the chunks
    :type chunk_size: int
    :returns: Iterator that returns list of length chunk_size until original lst is depleted.
    :rtype: Iterable
    """
    for i in range(0, len(lst), chunk_size):
        yield lst[i: i + chunk_size]


class Empty:
    """
    Represents an empty value.

    In edit functions we want to be able to edit values of model properties, e.g. the due_date of a process.
    Consider the following example:

        def edit(self, start_date, due_date)
            self.start_date = start_date
            self.due_date = due_date
            self.save()

    This OK, we can edit both the start and due date. But what if we want to be able to edit one of the two dates
    without providing or touching the other.

    We would have to change it as follows:

        def edit(self, start_date=None, due_date=None)
            if start_date:
                self.start_date = start_date
            if due_date:
                self.due_date = due_date
            self.save()

    Now, if we provide a start_date but not a due_date, the due_date would automatically get a None value. This way
    only the non-None values are edited and saved. Looks OK.

    But what if we want to empty a date and set it to null(None) in the database? If we we send None as a due_date value
    it would not get saved due to the None value checker we implemented in order to have optional parameters.

    Here comes the Empty class into play!

    The Empty class is just a Python class which does absolutely nothing and only serves the purpose of being unique.
    If we give the parameters in our edit function a default value which is an instance of the Empty class, we can check
    on that in order to know if we want to save the value. Like this:

        def edit(self, start_date=Empty(), due_date=Empty())
            if not isinstance(start_date, Empty):
                self.start_date = start_date
            if not isinstance(due_date, Empty):
                self.due_date = due_date
            self.save()

    Now both start_date and due_date are optional but they can also hold a None value which will lead to an actual
    edit.

    Happy coding!
    """

    __instance = None

    def __new__(cls, *args, **kwargs):
        """Only create a single instance of this class, i.e. simple singleton pattern."""
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance


empty = Empty()


def clean_empty_values(update_dict: Dict) -> Dict:
    """
    Clean a dictionary by removing all the keys that have an empty value.

    :param update_dict: dictionary with keys=request parameters
    :type update_dict: dict

    :return: same dictionary but cleaned up if there are values None
    :rtype: dict
    """
    cleaned_up_dict = {k: v for k, v in update_dict.items() if not isinstance(v, Empty)}
    return cleaned_up_dict


def get_offset_from_user_timezone(user: "User") -> int:
    """
    Retrieve the offset in minutes from UTC time compared to the user defined timezone.

    :param user: (optional) User used to calculate the offset in minutes.
    :type user: User object
    :return: number of minutes to the nearest integer
    """
    user_timezone = pytz.timezone(user.timezone.zone)
    user_time = datetime.now(user_timezone)
    offset = -int(user_time.tzinfo.utcoffset(user_time).total_seconds() / 60.0)
    return offset



from ssl import SSLError

from urllib3 import Retry
from urllib3.exceptions import MaxRetryError


class PykeRetry(Retry):
    """
    Pykechain Implementation of urllib3.Retry function.

    It contains a fast bailout of any SSLCertificate Errors.
    """

    def increment(
        self,
        method=None,
        url=None,
        response=None,
        error=None,
        _pool=None,
        _stacktrace=None,
    ):
        """In case of failed verification of self signed certificate we short circuit the retry routine."""
        if self._is_ssl_error(error):
            raise MaxRetryError(_pool, url, error)

        return super().increment(
            method=method,
            url=url,
            response=response,
            error=error,
            _pool=_pool,
            _stacktrace=_stacktrace,
        )

    def _is_ssl_error(self, error):
        return error and isinstance(error, SSLError)



from unittest import TestCase

from pykechain import Client
from pykechain.models.base import Base


class TestBase(TestCase):

    json = {"id": "123", "name": "test"}

    def test_id(self):
        obj = Base(self.json, None)

        self.assertEqual(obj.id, "123")

    def test_name(self):
        obj = Base(self.json, None)

        self.assertEqual(obj.name, "test")

    def test_given_client(self):
        client = Client()

        obj = Base(self.json, client)

        self.assertEqual(obj._client, client)
        self.assertIsInstance(obj._client, Client)



import datetime
import os
from typing import Dict, Optional

import requests

from pykechain.exceptions import APIError
from pykechain.models import Base
from pykechain.models.input_checks import check_type
from pykechain.utils import empty, clean_empty_values


class ExpiringDownload(Base):
    """Expiring Download class."""

    def __init__(self, json: Dict, **kwargs) -> None:
        """
        Init function.

        :param json: the json response to construct the :class:`ExpiringDownload` from
        :type json: dict
        """
        super().__init__(json, **kwargs)
        self.filename = json.get("content_file_name")
        self.expires_in = json.get("expires_in")
        self.expires_at = json.get("expires_at")
        self.token_hint = json.get("token_hint")

    def __repr__(self):  # pragma: no cover
        return f"<pyke ExpiringDownload id {self.id[-8:]}>"

    def save_as(self, target_dir: Optional[str] = None) -> None:
        """
        Save the Expiring Download content.

        :param target_dir: the target directory where the file will be stored
        :type target_dir: str
        """
        full_path = os.path.join(target_dir or os.getcwd(), self.filename)

        url = self._client._build_url("expiring_download_download", download_id=self.id)
        response = self._client._request("GET", url)
        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(
                f"Could not download file from Expiring download {self}", response=response
            )

        with open(full_path, "w+b") as f:
            for chunk in response:
                f.write(chunk)

    def delete(self) -> None:
        """Delete this expiring download.

        :raises APIError: if delete was not successful.
        """
        response = self._client._request(
            "DELETE", self._client._build_url("expiring_download", download_id=self.id)
        )

        if response.status_code != requests.codes.no_content:  # pragma: no cover
            raise APIError(f"Could not delete Expiring Download {self}", response=response)

    def edit(
        self,
        expires_at: Optional[datetime.datetime] = empty,
        expires_in: Optional[int] = empty,
        **kwargs,
    ) -> None:
        """
        Edit Expiring Download details.

        :param expires_at: The moment at which the ExpiringDownload will expire
        :type expires_at: datetime.datetime
        :param expires_in: The amount of time (in seconds) in which the ExpiringDownload will expire
        :type expires_in: int
        """
        update_dict = {
            "id": self.id,
            "expires_at": check_type(expires_at, datetime.datetime, "expires_at")
            or self.expires_at,
            "expires_in": check_type(expires_in, int, "expires_in") or self.expires_in,
        }

        if kwargs:  # pragma: no cover
            update_dict.update(**kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)

        response = self._client._request(
            "PUT",
            self._client._build_url("expiring_download", download_id=self.id),
            json=update_dict,
        )

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Expiring Download {self}", response=response)

        self.refresh(json=response.json()["results"][0])

    def upload(self, content_path):
        """
        Upload a file to the Expiring Download.

        .. versionadded:: 3.10.0

        :param content_path: path to the file to upload.
        :type content_path: basestring
        :raises APIError: if the file could not be uploaded.
        :raises OSError: if the file could not be located on disk.
        """
        if os.path.exists(content_path):
            self._upload(content_path=content_path)
        else:
            raise OSError(f"Could not locate file to upload in '{content_path}'")

    def _upload(self, content_path):
        url = self._client._build_url("expiring_download_upload", download_id=self.id)

        with open(content_path, "rb") as file:
            response = self._client._request(
                "POST", url, files={"attachment": (os.path.basename(content_path), file)}
            )

        if response.status_code not in (
            requests.codes.accepted,
            requests.codes.ok,
        ):  # pragma: no cover
            raise APIError(
                f"Could not upload  file to Expiring Download {self}", response=response
            )

        self.refresh(json=response.json()["results"][0])



from pykechain.exceptions import _DeprecationMixin
from pykechain.models import DatetimeProperty


class DatetimeProperty2(DatetimeProperty, _DeprecationMixin):
    """A virtual object representing a KE-chain datetime property."""

    pass



from typing import Any, List, Dict, Optional, Union, Tuple, Iterable, TypeVar

import requests
from jsonschema import validate

from pykechain.defaults import API_EXTRA_PARAMS
from pykechain.enums import Category
from pykechain.exceptions import APIError, IllegalArgumentError
from pykechain.models import Base, BaseInScope
from pykechain.models.input_checks import check_text, check_type
from pykechain.models.representations.component import RepresentationsComponent
from pykechain.models.validators import PropertyValidator
from pykechain.models.validators.validator_schemas import options_json_schema
from pykechain.utils import clean_empty_values, empty

T = TypeVar("T")


class Property(BaseInScope):
    """A virtual object representing a KE-chain property.

    .. versionadded: 3.0
       This is a `Property` to communicate with a KE-chain 3 backend.

    :cvar bulk_update: flag to postpone update of properties until manually requested
    :type bulk_update: bool
    :ivar type: The property type of the property. One of the types described in :class:`pykechain.enums.PropertyType`
    :type type: str
    :ivar category: The category of the property, either `Category.MODEL` of `Category.INSTANCE`
    :type category: str
    :ivar description: description of the property
    :type description: str or None
    :ivar unit: unit of measure of the property
    :type unit: str or None
    :ivar model: the id of the model (not the model object)
    :type model: str
    :ivar output: a boolean if the value is configured as an output (in an activity)
    :type output: bool
    :ivar part: The (parent) part in which this property is available
    :type part: :class:`Part`
    :ivar value: the property value, can be set as well as property
    :type value: Any
    :ivar validators: the list of validators that are available in the property
    :type validators: List[PropertyValidator]
    :ivar is_valid: if the property conforms to the validators
    :type is_valid: bool
    :ivar is_invalid: if the property does not conform to the validator
    :type is_invalid: bool
    """

    _USE_BULK_UPDATE = False
    _update_package = dict()

    def __init__(self, json, **kwargs):
        """Construct a Property from a json object."""
        super().__init__(json, **kwargs)

        self.output: bool = json.get("output")
        self.model_id: Optional[str] = json.get("model_id")
        self.part_id = json.get("part_id")
        self.ref = json.get("ref")
        self.type = json.get("property_type")
        self.category = json.get("category")
        self.description = json.get("description", None)
        self.unit = json.get("unit", None)
        self.order = json.get("order")

        # Create protected variables
        self._value: Any = json.get("value")
        self._options: Dict = json.get("value_options", {})
        self._part: Optional["Part"] = None
        self._model: Optional["Property"] = None
        self._validators: List[PropertyValidator] = []
        self._validation_results: List = []
        self._validation_reasons: List = []

        self._representations_container = RepresentationsComponent(
            self,
            self._options.get("representations", {}),
            self._save_representations,
        )

        if "validators" in self._options:
            self._parse_validators()

    def _options_valid(self) -> bool:
        """Validate the options of the Property object.

        It will validate if the Property options are valid against the options JSON schema.

        :raises jsonschema.exceptions.ValidationError: if the options value is invalid
        :raises jsonschema.exceptions.SchemaError: if the JSON schema of the options is invalid
        :returns: Boolean True if valid
        """
        validate(self._options, options_json_schema)
        return True

    def refresh(
        self, json: Optional[Dict] = None, url: Optional[str] = None, extra_params: Optional = None
    ) -> None:
        """Refresh the object in place."""
        super().refresh(
            json=json,
            url=self._client._build_url("property", property_id=self.id),
            extra_params=API_EXTRA_PARAMS["property"],
        )

    def has_value(self) -> bool:
        """Predicate to indicate if the property has a value set.

        This predicate determines if the property has a value set. It will not make a call to KE-chain API (in case
        of reference properties). So it is a tiny fraction 'cheaper' in terms of processing time than checking the
        `Property.value` itself.

        It will return True if the property_type is a Boolean and set to a value of False.

        :returns: True if the property has a value set, otherwise (also when value is None) returns False
        :rtype: Bool
        """
        if isinstance(self._value, (float, int, bool)):
            return True  # to prevent "bool(0.00) = False" or "False = False"
        else:
            return bool(self._value)

    @property
    def use_bulk_update(self):
        """Set or get the toggle to asynchronously update property values."""
        # set the class attribute to make this value a singleton
        return self.__class__._USE_BULK_UPDATE

    @use_bulk_update.setter
    def use_bulk_update(self, value):
        self.__class__.set_bulk_update(value)

    @classmethod
    def set_bulk_update(cls, value):
        """Set global class attribute to toggle the use of bulk-updates of properties."""
        assert isinstance(
            value, bool
        ), f"`bulk_update` must be set to a boolean, not {type(value)}"
        cls._USE_BULK_UPDATE = value

    @property
    def value(self) -> Any:
        """Retrieve the data value of a property.

        Setting this value will immediately update the property in KE-chain.

        :returns: the value
        """
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        value = self.serialize_value(value)

        # check if the options are valid
        self._options_valid()

        if self.use_bulk_update:
            self._pend_update(dict(value=value))
            self._value = value
        else:
            self._put_value(value)

    @classmethod
    def update_values(cls, client: "Client", use_bulk_update: bool = False) -> None:
        """
        Perform the bulk update of property values using the stored values in the `Property` class.

        :param client: Client object
        :type client: Client
        :param use_bulk_update: set the class attribute, defaults to False.
        :type use_bulk_update: bool
        :return: None
        """
        if cls._USE_BULK_UPDATE:
            properties = [dict(id=key, **values) for key, values in cls._update_package.items()]
            client.update_properties(properties=properties)
            cls._update_package = dict()
        cls.set_bulk_update(use_bulk_update)

    def _pend_update(self, data):
        """Store the value to be send at a later point in time using `update_values`."""
        existing_data = self.__class__._update_package.get(self.id, None)
        if existing_data:
            existing_data.update(data)
        else:
            self.__class__._update_package[self.id] = data

    def _put_value(self, value):
        """Send the value to KE-chain."""
        url = self._client._build_url("property", property_id=self.id)

        response = self._client._request(
            "PUT", url, params=API_EXTRA_PARAMS["property"], json={"value": value}
        )

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Property {self}", response=response)

        self.refresh(json=response.json()["results"][0])

    def serialize_value(self, value: [T]) -> T:
        """
        Serialize the value to be set on the property.

        :param value: non-serialized value
        :type value: Any
        :return: serialized value
        """
        return value.id if isinstance(value, Base) else value

    @property
    def part(self) -> "Part":
        """
        Retrieve the part that holds this Property.

        :returns: The :class:`Part` associated to this property
        :raises APIError: if the `Part` is not found
        """
        if self._part is None:
            self._part = self._client.part(pk=self.part_id, category=self.category)
        return self._part

    def model(self) -> "AnyProperty":
        """
        Model object of the property if the property is an instance otherwise itself.

        Will cache the model object in order to not generate too many API calls. Otherwise will make an API call
        to the backend to retrieve its model object.

        :return: `Property` model object if the current `Property` is an instance.
        :rtype: :class:`pykechain.models.AnyProperty`
        """
        if self.category == Category.MODEL:
            return self
        elif self._model is None:
            self._model = self._client.property(pk=self.model_id, category=Category.MODEL)
        return self._model

    @property
    def validators(self):
        """Provide list of Validator objects.

        :returns: list of :class:`PropertyValidator` objects
        :rtype: list(PropertyValidator)
        """
        return self._validators

    @validators.setter
    def validators(self, validators: Iterable[PropertyValidator]) -> None:
        if self.category != Category.MODEL:
            raise IllegalArgumentError(
                "To update the list of validators, it can only work on "
                "`Property` of category 'MODEL'"
            )

        if not isinstance(validators, (tuple, list)) or not all(
            isinstance(v, PropertyValidator) for v in validators
        ):
            raise IllegalArgumentError(
                "Should be a list or tuple with PropertyValidator objects, got {}".format(
                    type(validators)
                )
            )
        for validator in validators:
            validator.validate_json()

        # set the internal validators list
        self._validators = list(set(validators))

        # dump to _json options
        self._dump_validators()

        # update the options to KE-chain backend
        self.edit(options=self._options)

    def _parse_validators(self):
        """Parse the validator in the options to validators."""
        self._validators = []
        validators_json = self._options.get("validators")
        for validator_json in validators_json:
            self._validators.append(PropertyValidator.parse(json=validator_json))

    def _dump_validators(self):
        """Dump the validators as json inside the _options dictionary with the key `validators`."""
        validators_json = []
        for validator in self._validators:
            if isinstance(validator, PropertyValidator):
                validators_json.append(validator.as_json())
            else:
                raise APIError(f"validator is not a PropertyValidator: '{validator}'")
        if self._options.get("validators", list()) == validators_json:
            # no change
            pass
        else:
            new_options = self._options.copy()  # make a copy
            new_options.update({"validators": validators_json})
            validate(new_options, options_json_schema)
            self._options = new_options

    @property
    def is_valid(self) -> Optional[bool]:
        """Determine if the value in the property is valid.

        If the value of the property is validated as 'valid', than returns a True, otherwise a False.
        When no validators are configured, returns a None. It checks against all configured validators
        and returns a single boolean outcome.

        :returns: True when the `value` is valid
        :rtype: bool or None
        """
        if not self._validators:
            return None
        else:
            self.validate(reason=False)
            if all([vr is None for vr in self._validation_results]):
                return None
            else:
                return all(self._validation_results)

    @property
    def is_invalid(self) -> Optional[bool]:
        """Determine if the value in the property is invalid.

        If the value of the property is validated as 'invalid', than returns a True, otherwise a False.
        When no validators are configured, returns a None. It checks against all configured validators
        and returns a single boolean outcome.

        :returns: True when the `value` is invalid
        :rtype: bool
        """
        return not self.is_valid if self.is_valid is not None else None

    def validate(self, reason: bool = True) -> List[Union[bool, Tuple]]:
        """Return the validation results and include an (optional) reason.

        If reason keyword is true, the validation is returned for each validation
        the [(<result: bool>, <reason:str>), ...]. If reason is False, only a single list of validation results
        for each configured validator is returned.

        :param reason: (optional) switch to indicate if the reason of the validation should be provided
        :type reason: bool
        :return: list of validation results [bool, bool, ...] or
                 a list of validation results, reasons [(bool, str), ...]
        :rtype: list(bool) or list((bool, str))
        :raises Exception: for incorrect validators or incompatible values
        """
        self._validation_results = [
            validator.is_valid(self._value) for validator in self._validators
        ]
        self._validation_reasons = [validator.get_reason() for validator in self._validators]

        if reason:
            return list(zip(self._validation_results, self._validation_reasons))
        else:
            return self._validation_results

    @property
    def representations(self):
        """Get and set the property representations."""
        return self._representations_container.get_representations()

    @representations.setter
    def representations(self, value):
        if self.category != Category.MODEL:
            raise IllegalArgumentError(
                "To update the list of representations, it can only work on a "
                "`Property` of category '{}'".format(Category.MODEL)
            )

        self._representations_container.set_representations(value)

    def _save_representations(self, representation_options):
        self._options.update({"representations": representation_options})
        self.edit(options=self._options)

    @classmethod
    def create(cls, json: dict, **kwargs) -> "AnyProperty":
        """Create a property based on the json data.

        This method will attach the right class to a property, enabling the use of type-specific methods.

        It does not create a property object in KE-chain. But a pseudo :class:`Property` object.

        :param json: the json from which the :class:`Property` object to create
        :type json: dict
        :return: a :class:`Property` object
        """
        property_type = json.get("property_type")

        from pykechain.models import property_type_to_class_map

        # Get specific Property subclass, defaulting to Property itself
        property_class = property_type_to_class_map.get(property_type, Property)

        # Call constructor and return new object
        return property_class(json, **kwargs)

    def edit(
        self,
        name: Optional[str] = empty,
        description: Optional[str] = empty,
        unit: Optional[str] = empty,
        options: Optional[Dict] = empty,
        **kwargs,
    ) -> None:
        """Edit the details of a property (model).

        Setting an input to None will clear out the value (exception being name).

        :param name: (optional) new name of the property to edit. Cannot be cleared.
        :type name: basestring or None or Empty
        :param description: (optional) new description of the property. Can be cleared.
        :type description: basestring or None or Empty
        :param unit: (optional) new unit of the property. Can be cleared.
        :type unit: basestring or None or Empty
        :param options: (options) new options of the property. Can be cleared.
        :type options: dict or None or Empty
        :param kwargs: (optional) additional kwargs to be edited
        :return: None
        :raises APIError: When unable to edit the property
        :raises IllegalArgumentError: when the type of the input is provided incorrect.

        Examples
        --------
        >>> front_fork = project.part('Front Fork')
        >>> color_property = front_fork.property(name='Color')
        >>> color_property.edit(name='Shade',description='Could also be called tint, depending on mixture',unit='RGB')

        >>> wheel_property_reference = self.project.model('Bike').property('Reference wheel')
        >>> wheel_model = self.project.model('Wheel')
        >>> diameter_property = wheel_model.property('Diameter')
        >>> spokes_property = wheel_model.property('Spokes')
        >>> prefilters = {'property_value': diameter_property.id + ":{}:lte".format(15)}
        >>> propmodels_excl = [spokes_property.id]
        >>> options = dict()
        >>> options['prefilters'] = prefilters
        >>> options['propmodels_excl'] = propmodels_excl
        >>> wheel_property_reference.edit(options=options)

        Not mentioning an input parameter in the function will leave it unchanged. Setting a parameter as None will
        clear its value (where that is possible). The example below will clear the description, but leave everything
        else unchanged.

        >>> wheel_property.edit(description=None)

        """
        update_dict = {
            "name": check_text(name, "name") or self.name,
            "description": check_text(description, "description") or "",
            "unit": check_text(unit, "unit") or "",
            "value_options": check_type(options, dict, "options") or dict(),
        }

        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)

        if self.use_bulk_update:
            self._pend_update(data=update_dict)
        else:
            update_dict["id"] = self.id

            response = self._client._request(
                "PUT",
                self._client._build_url("property", property_id=self.id),
                params=API_EXTRA_PARAMS["property"],
                json=update_dict,
            )

            if response.status_code != requests.codes.ok:  # pragma: no cover
                raise APIError(f"Could not update Property {self}", response=response)

            self.refresh(json=response.json()["results"][0])

    def delete(self) -> None:
        """Delete this property.

        :return: None
        :raises APIError: if delete was not successful
        """
        response = self._client._request(
            "DELETE", self._client._build_url("property", property_id=self.id)
        )

        if response.status_code != requests.codes.no_content:  # pragma: no cover
            raise APIError(f"Could not delete Property {self}", response=response)

    def copy(self, target_part: "Part", name: Optional[str] = None) -> "Property":
        """Copy a property model or instance.

        :param target_part: `Part` object under which the desired `Property` is copied
        :type target_part: :class:`Part`
        :param name: how the copied `Property` should be called
        :type name: basestring
        :return: copied :class:`Property` model.
        :raises IllegalArgumentError: if property and target_part have different `Category`

        Example
        -------
        >>> property_to_copy = client.property(name='Diameter')
        >>> bike = client.model('Bike')
        >>> property_to_copy.copy(target_part=bike, name='Bike diameter?')

        """
        from pykechain.models import Part

        check_type(target_part, Part, "target_part")

        name = check_text(name, "name") or self.name
        if self.category == Category.MODEL and target_part.category == Category.MODEL:
            # Cannot move a `Property` model under a `Part` instance or vice versa
            copied_property_model = target_part.add_property(
                name=name,
                property_type=self.type,
                description=self.description,
                unit=self.unit,
                default_value=self.value,
                options=self._options,
            )
            return copied_property_model
        elif self.category == Category.INSTANCE and target_part.category == Category.INSTANCE:
            target_model = target_part.model()
            self_model = self.model()
            target_model.add_property(
                name=name,
                property_type=self_model.type,
                description=self_model.description,
                unit=self_model.unit,
                default_value=self_model.value,
                options=self_model._options,
            )
            target_part.refresh()
            copied_property_instance = target_part.property(name=name)
            copied_property_instance.value = self.value
            return copied_property_instance
        else:
            raise IllegalArgumentError(
                'property "{}" and target part "{}" must have the same category'.format(
                    self.name, target_part.name
                )
            )

    def move(self, target_part: "Part", name: Optional[str] = None) -> "Property":
        """Move a property model or instance.

        :param target_part: `Part` object under which the desired `Property` is moved
        :type target_part: :class:`Part`
        :param name: how the moved `Property` should be called
        :type name: basestring
        :return: copied :class:`Property` model.
        :raises IllegalArgumentError: if property and target_part have different `Category`

        Example
        -------
        >>> property_to_move = client.property(name='Diameter')
        >>> bike = client.model('Bike')
        >>> property_to_move.move(target_part=bike, name='Bike diameter?')

        """
        moved_property = self.copy(target_part=target_part, name=name)

        if self.category == Category.MODEL:
            self.delete()
        else:
            self.model().delete()
        return moved_property



from unittest import TestCase

from requests import PreparedRequest

from pykechain.exceptions import (
    APIError,
    ForbiddenError,
    MultipleFoundError,
    NotFoundError,
    ClientError,
    IllegalArgumentError,
    InspectorComponentError,
)
from tests.classes import TestBetamax


class TestExceptions(TestCase):
    def test_api_error(self):
        api_error = APIError()

        with self.assertRaises(Exception):
            raise api_error

    def test_api_error_with_message(self):
        api_error = APIError("This is the error message.")

        with self.assertRaises(APIError):
            raise api_error

    def test_api_error_with_objects(self):
        APIError(["text in a list"])
        APIError(("text in a tuple",))
        APIError(dict(text="in a dict"))

    def test_inheritance(self):
        self.assertTrue(issubclass(APIError, BaseException))
        self.assertTrue(issubclass(ForbiddenError, APIError))
        self.assertTrue(issubclass(MultipleFoundError, APIError))
        self.assertTrue(issubclass(NotFoundError, APIError))
        self.assertTrue(issubclass(ClientError, APIError))

        self.assertFalse(issubclass(IllegalArgumentError, APIError))
        self.assertFalse(issubclass(InspectorComponentError, APIError))

    def test_creation(self):
        for test_class in [
            APIError,
            ForbiddenError,
            MultipleFoundError,
            NotFoundError,
            ClientError,
            IllegalArgumentError,
            InspectorComponentError,
        ]:
            with self.subTest(msg=test_class.__name__):
                test_class()


class TestExceptionsLive(TestBetamax):
    def setUp(self):
        super().setUp()
        url = self.client._build_url("activities")
        self.response = self.client._request(method="GET", url=url, params={"limit": 1})

    def test_api_error_with_response(self):
        api_error = APIError(response=self.response)

        # testing
        self.assertIsInstance(api_error.request, PreparedRequest)
        self.assertEqual(self.response, api_error.response)
        with self.assertRaises(APIError):
            raise api_error

    def test_api_error_with_argument_and_response(self):
        APIError(dict(text="my dict"), response=self.response)



import mimetypes
import re
from typing import (
    Any,
    Union,
    Tuple,
    Optional,
    Dict,
    List,
)  # noqa: F401 # pylint: disable=unused-import

from pykechain.enums import PropertyVTypes
from pykechain.models.validators.mime_types_defaults import predefined_mimes
from pykechain.models.validators.validator_schemas import (
    filesizevalidator_schema,
    fileextensionvalidator_schema,
)
from pykechain.models.validators.validators_base import PropertyValidator
from pykechain.utils import EMAIL_REGEX_PATTERN


class NumericRangeValidator(PropertyValidator):
    """
    A numeric range validator, which validates a number between a range.

    The range validates positively upto and **including** the minvalue and maxvalue.

    An added ability is the check if the number conforms to a step within that range.
    The validation checks for both integer and floats. The stepsize is only enforced when the :attr:`enforce_stepsize`
    is set to `True`. This enforcement is accurate to an accuracy set in the :const:`.accuracy`
    (normally set to be 1E-6).

    .. versionadded:: 2.2

    :ivar minvalue: minimum value of the range
    :type minvalue: float or int
    :ivar maxvalue: maximum value of the range
    :type maxvalue: float or int
    :ivar stepsize: stepsize
    :type stepsize: float or int
    :ivar enforce_stepsize: flag to ensure that the stepsize is enforced
    :type enforce_stepsize: bool

    Examples
    --------
    >>> validator = NumericRangeValidator(minvalue=0, maxvalue=50)
    >>> validator.is_valid(42)
    True
    >>> validator.is_valid(50)
    True
    >>> validator.is_valid(50.0001)
    False
    >>> validator.is_valid(-1)
    False
    >>> validator.get_reason()
    Value '-1' should be between 0 and 50

    >>> stepper = NumericRangeValidator(stepsize=1000, enforce_stepsize=True)
    >>> stepper.is_valid(2000)
    True

    """

    vtype = PropertyVTypes.NUMERICRANGE

    def __init__(
        self,
        json=None,
        minvalue=None,
        maxvalue=None,
        stepsize=None,
        enforce_stepsize=None,
        **kwargs,
    ):
        """Construct the numeric range validator."""
        super().__init__(json=json, **kwargs)

        if minvalue is not None:
            self._config["minvalue"] = minvalue
        if maxvalue is not None:
            self._config["maxvalue"] = maxvalue
        if stepsize is not None:
            self._config["stepsize"] = stepsize
        if enforce_stepsize is not None:
            self._config["enforce_stepsize"] = enforce_stepsize

        if self._config.get("minvalue") is None:
            self.minvalue = float("-inf")
        else:
            self.minvalue = self._config.get("minvalue")
        if self._config.get("maxvalue") is None:
            self.maxvalue = float("inf")
        else:
            self.maxvalue = self._config.get("maxvalue")

        self.stepsize = self._config.get("stepsize", None)
        self.enforce_stepsize = self._config.get("enforce_stepsize", None)

        if self.minvalue > self.maxvalue:
            raise Exception(
                "The minvalue ({}) should be smaller than the maxvalue ({}) of the numeric "
                "range validation".format(self.minvalue, self.maxvalue)
            )
        if self.enforce_stepsize and self.stepsize is None:
            raise Exception("The stepsize should be provided when enforcing stepsize")

    def _logic(self, value: Any = None) -> Tuple[Union[bool, None], str]:
        basereason = f"Value '{value}' should be between {self.minvalue} and {self.maxvalue}"
        self._validation_result, self._validation_reason = None, "No reason"

        if value is not None:
            self._validation_result = value >= self.minvalue and value <= self.maxvalue
            if not self._validation_result:
                self._validation_reason = basereason
            else:
                self._validation_reason = basereason.replace("should be", "is")

        if self.stepsize != 1 and self.enforce_stepsize:
            # to account also for floating point stepsize checks: https://stackoverflow.com/a/30445184/246235
            if self.minvalue == float("-inf"):
                self._validation_result = (
                    abs(value / self.stepsize - round(value / self.stepsize)) < self.accuracy
                )
            else:
                self._validation_result = (
                    abs(
                        (value - self.minvalue) / self.stepsize
                        - round((value - self.minvalue) / self.stepsize)
                    )
                    < self.accuracy
                )

            if not self._validation_result:
                self._validation_reason = (
                    "Value '{}' is not in alignment with a stepsize of {}".format(
                        value, self.stepsize
                    )
                )

        return self._validation_result, self._validation_reason


class RequiredFieldValidator(PropertyValidator):
    """
    Required field validator ensures that a value is provided.

    Does validate all values. Does not validate `None` or `''` (empty string).

    .. versionadded:: 2.2

    Examples
    --------
    >>> validator = RequiredFieldValidator()
    >>> validator.is_valid("A value")
    True
    >>> validator.is_valid("")
    False
    >>> validator.is_valid(None)
    False
    >>> validator.get_reason()
    "Value is required"

    """

    vtype = PropertyVTypes.REQUIREDFIELD

    def _logic(self, value: Any = None) -> Tuple[Union[bool, None], str]:
        basereason = "Value is required"
        self._validation_result, self._validation_reason = None, "No reason"

        if (
            value is not None
            and value != ""
            and value != list()
            and value != tuple()
            and value != set()
        ):
            self._validation_result = True
            self._validation_reason = "Value is provided"
        else:
            self._validation_result = False
            self._validation_reason = basereason

        return self._validation_result, self._validation_reason


class BooleanFieldValidator(PropertyValidator):
    """A boolean field validator.

    This is a stub implementation. Should validate if a value is either 'truthy' or 'falsy'.
    """

    vtype = PropertyVTypes.BOOLEANFIELD


class EvenNumberValidator(PropertyValidator):
    """An even number validator that validates `True` when the number is even.

    Even numbers are scalar numbers which can be diveded by 2 and return a scalar. Floating point numbers are converted
    to integer first. So `int(4.5)` = 4.

    .. versionadded:: 2.2

    Example
    -------
    >>> validator = EvenNumberValidator()
    >>> validator.is_valid(4)
    True
    >>> validator.is_valid(4.5)  # float is converted to integer first
    True

    """

    vtype = PropertyVTypes.EVENNUMBER

    def _logic(self, value: Any = None) -> Tuple[Union[bool, None], str]:
        if value is None:
            self._validation_result, self.validation_reason = None, "No reason"
            return self._validation_result, self._validation_reason
        if not isinstance(value, (int, float)):
            self._validation_result, self.validation_reason = (
                False,
                "Value should be an integer, or float (floored)",
            )
            return self._validation_result, self._validation_reason

        basereason = f"Value '{value}' should be an even number"

        self._validation_result = int(value) % 2 < self.accuracy
        if self._validation_result:
            self._validation_reason = basereason.replace("should be", "is")
            return self._validation_result, self._validation_reason
        else:
            self._validation_reason = basereason
            return self._validation_result, self._validation_reason


class OddNumberValidator(PropertyValidator):
    """A odd number validator that validates `True` when the number is odd.

    Example
    -------
    >>> validator = OddNumberValidator()
    >>> validator.is_valid(3)
    True
    >>> validator.is_valid(3.5)  # float is converted to integer first
    True

    """

    vtype = PropertyVTypes.ODDNUMBER

    def _logic(self, value: Any = None) -> Tuple[Union[bool, None], str]:
        if value is None:
            self._validation_result, self.validation_reason = None, "No reason"
            return self._validation_result, self._validation_reason
        if not isinstance(value, (int, float)):
            self._validation_result, self.validation_reason = (
                False,
                "Value should be an integer, or float (floored)",
            )
            return self._validation_result, self._validation_reason

        basereason = f"Value '{value}' should be a odd number"

        self._validation_result = int(value) % 2 > self.accuracy
        if self._validation_result:
            self._validation_reason = basereason.replace("should be", "is")
            return self._validation_result, self._validation_reason
        else:
            self._validation_reason = basereason
            return self._validation_result, self._validation_reason


class SingleReferenceValidator(PropertyValidator):
    """A single reference validator, ensuring that only a single reference is selected.

    .. versionadded:: 2.2
    """

    vtype = PropertyVTypes.SINGLEREFERENCE

    def _logic(self, value: Any = None) -> Tuple[Union[bool, None], str]:
        if value is None:
            self._validation_result, self.validation_reason = None, "No reason"
            return self._validation_result, self._validation_reason
        if not isinstance(value, (list, tuple, set)):
            self._validation_result, self.validation_reason = (
                False,
                "Value should be a list, tuple or set",
            )
            return self._validation_result, self._validation_reason

        self._validation_result = len(value) == 1 or len(value) == 0
        if self._validation_result:
            self._validation_reason = "A single or no value is selected"
            return self._validation_result, self._validation_reason
        else:
            self._validation_reason = "More than a single instance is selected"
            return self._validation_result, self._validation_reason


class RegexStringValidator(PropertyValidator):
    """
    A regular expression string validator.

    With a configured regex pattern, a string value is compared and matched against this pattern. If there is
    a positive match, the validator validates correctly.

    For more information on constructing regex strings, see the `python documentation`_, `regex101.com`_, or
    `regexr.com`_.

    .. versionadded:: 2.2

    :ivar pattern: the regex pattern to which the provided value is matched against.

    Example
    -------
    >>> validator = RegexStringValidator(pattern=r"Yes|Y|1|Ok")
    >>> validator.is_valid("Yes")
    True
    >>> validator.is_valid("No")
    False

    .. _python documentation: https://docs.python.org/2/library/re.html
    .. _regex101.com: https://regex101.com/
    .. _regexr.com: https://regexr.com/

    """

    vtype = PropertyVTypes.REGEXSTRING

    def __init__(self, json=None, pattern=None, **kwargs):
        """Construct an regex string validator.

        If no pattern is provided than the regexstring `'.+'` will be used, which matches all provided text with
        at least a single character. Does not match `''` (empty string).

        :param json: (optional) dict (json) object to construct the object from
        :type json: dict
        :param pattern: (optional) valid regex string, defaults to r'.+' which matches all text.
        :type text: basestring
        :param kwargs: (optional) additional kwargs to pass down
        :type kwargs: dict
        """
        super().__init__(json=json, **kwargs)

        if pattern is not None:
            self._config["pattern"] = pattern

        self.pattern = self._config.get("pattern", r".+")
        self._re = re.compile(self.pattern)

    def _logic(self, value: Any = None) -> Tuple[Union[bool, None], str]:
        if value is None:
            self._validation_result, self.validation_reason = None, "No reason"
            return self._validation_result, self._validation_reason

        basereason = f"Value '{value}' should match the regex pattern '{self.pattern}'"

        self._validation_result = re.match(self._re, value) is not None
        if not self._validation_result:
            self._validation_reason = basereason
        else:
            self._validation_reason = basereason.replace("should match", "matches")

        return self._validation_result, self._validation_reason


class EmailValidator(RegexStringValidator):
    """
    A email string validator.

    :cvar pattern: the email regex pattern to which the provided value is matched against.
    """

    pattern = EMAIL_REGEX_PATTERN

    def __init__(self, json=None, **kwargs):
        """Construct an email string validator.

        :param json: (optional) dict (json) object to construct the object from
        :type json: dict
        :param kwargs: (optional) additional kwargs to pass down
        :type kwargs: dict
        """
        super().__init__(json=json, pattern=self.pattern, **kwargs)


class AlwaysAllowValidator(PropertyValidator):
    """
    An always allow Validator.

    Will always return True.
    """

    vtype = PropertyVTypes.ALWAYSALLOW

    def _logic(self, value: Any = None) -> Tuple[Union[bool, None], str]:
        """Process the inner logic of the validator.

        The validation results are returned as tuple (boolean (true/false), reasontext)
        """
        return True, "Always True"


class FileSizeValidator(PropertyValidator):
    """A file size Validator.

    The actual size of the file cannot be checked in pykechain without downloading this from the server, hence
    when the validator is used inside an attachment property, the validator returns always being valid.

    :ivar max_size: maximum size to check
    :type max_size: Union[int,float]

    Example
    -------
    >>> validator = FileSizeValidator(max_size=100)
    >>> validator.is_valid(100)
    True
    >>> validator.is_valid(-1)
    False
    >>> validator.is_valid("attachments/12345678-1234-5678-1234-567812345678/some_file.txt")
    True
    >>> validator.get_reason()
    We determine the filesize of 'some_file.txt' to be valid. We cannot check it at this end.

    """

    vtype = PropertyVTypes.FILESIZE
    jsonschema = filesizevalidator_schema

    def __init__(
        self, json: Optional[Dict] = None, max_size: Optional[Union[int, float]] = None, **kwargs
    ):
        """Construct a file size validator.

        :param json: (optional) dict (json) object to construct the object from
        :type json: Optional[Dict]
        :param max_size: (optional) number that counts as maximum size of the file
        :type accept: Optional[Union[int,float]]
        :param kwargs: (optional) additional kwargs to pass down
        """
        super().__init__(json=json, **kwargs)
        if max_size is not None:
            if isinstance(max_size, (int, float)):
                self._config["maxSize"] = int(max_size)
            else:
                raise ValueError("`max_size` should be a number.")
        self.max_size = self._config.get("maxSize", float("inf"))

    def _logic(
        self, value: Optional[Union[int, float]] = None
    ) -> Tuple[Optional[bool], Optional[str]]:
        """Based on a filesize (numeric) or  filepath of the property (value), the filesize is checked."""
        if value is None:
            return None, "No reason"

        basereason = f"Value '{value}' should be of a size less then '{self.max_size}'"

        if isinstance(value, (int, float)):
            if int(value) <= self.max_size and int(value) >= 0:
                return True, basereason.replace("should", "is")
            else:
                return False, basereason

        return (
            True,
            f"We determine the filesize of '{value}' to be valid. We cannot check it at this end.",
        )


class FileExtensionValidator(PropertyValidator):
    """A file extension Validator.

    It checks the value of the property attachment against a list of acceptable mime types or file extensions.

    Example
    -------
    >>> validator = FileExtensionValidator(accept=[".png", ".jpg"])
    >>> validator.is_valid("picture.jpg")
    True
    >>> validator.is_valid("document.pdf")
    False
    >>> validator.is_valid("attachments/12345678-1234-5678-1234-567812345678/some_file.txt")
    False

    >>> validator = FileExtensionValidator(accept=["application/pdf",
    ...                                            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"])
    >>> validator.is_valid("document.pdf")
    True
    >>> validator.is_valid("comma-separated-values.csv")
    False
    >>> validator.is_valid("attachments/12345678-1234-5678-1234-567812345678/modern_excel.xlsx")
    True

    """

    vtype = PropertyVTypes.FILEEXTENSION
    jsonschema = fileextensionvalidator_schema
    mimetype_regex = r"^[-\w.]+/[-\w.\*]+$"

    def __init__(
        self, json: Optional[Dict] = None, accept: Optional[Union[str, List[str]]] = None, **kwargs
    ):
        """Construct a file extension validator.

        :param json: (optional) dict (json) object to construct the object from
        :type json: Optional[Dict]
        :param accept: (optional) list of mimetypes or file extensions (including a `.`, eg `.csv`, `.pdf`)
        :type accept: Optional[List[Text]]
        :param kwargs: (optional) additional kwargs to pass down
        """
        super().__init__(json=json, **kwargs)
        if accept is not None:
            if isinstance(accept, str):
                self._config["accept"] = accept.split(",")
            elif isinstance(accept, List):
                self._config["accept"] = accept
            else:
                raise ValueError("`accept` should be a commaseparated list or a list of strings.")

        self.accept = self._config.get("accept", None)
        self._accepted_mimetypes = self._convert_to_mimetypes(self.accept)

    def _convert_to_mimetypes(self, accept: List[str]) -> Optional[List[str]]:
        """
        Convert accept array to array of mimetypes.

        1. convert accept list to array of mimetypes
        2. convert aggregator (if inside mime_types_defaults) to list of mimetypes

        :param accept: list of mimetypes or extensions
        :type accept: List[Text]
        :return: array of mimetypes:
        :rtype: List[Text]
        """
        if accept is None:
            return None

        marray = []
        for item in accept:
            # check if the item in the accept array is a mimetype on its own.
            if re.match(self.mimetype_regex, item):
                if item in predefined_mimes:
                    marray.extend(predefined_mimes.get(item))
                else:
                    marray.append(item)
            else:
                # we assume this is an extension.
                # we can only guess a url, we make a url like: "file.ext" to check.
                fake_filename = f"file{item}" if item.startswith(".") else f"file.{item}"

                # do guess
                guess, _ = mimetypes.guess_type(fake_filename)
                marray.append(guess) if guess is not None else print(guess)

        return marray

    def _logic(self, value: Optional[str] = None) -> Tuple[Optional[bool], Optional[str]]:
        """Based on the filename of the property (value), the type is checked.

        1. convert filename to mimetype
        3. check if the filename is inside the array of mimetypes. self._accepted_mimetypes
        """
        if value is None:
            return None, "No reason"

        basereason = f"Value '{value}' should match the mime types '{self.accept}'"

        guessed_type, _ = mimetypes.guess_type(value)
        if guessed_type is None:
            return False, f"Could not determine the mimetype of '{value}'"
        elif guessed_type in self._accepted_mimetypes:
            return True, basereason.replace("match", "matches")

        return False, basereason



from pykechain.exceptions import _DeprecationMixin
from pykechain.models import ActivityReferencesProperty


class ActivityReferencesProperty2(ActivityReferencesProperty, _DeprecationMixin):
    """A virtual object representing a KE-chain Activity References property.

    .. versionadded:: 3.7
    """

    pass



import datetime

import pytz

from .base import Base
from ..enums import LanguageCodes


class User(Base):
    """A virtual object representing a KE-chain user.

    :ivar username: username of the user
    :type username: str
    :ivar name: username of the user (compatibility)
    :type name: str
    :ivar id: userid of the user
    :type id: int
    :ivar timezone: timezone of the User (defaults to <UTC>)
    :type timezone: timezone object
    :ivar language: language of the User (defaults to 'en')
    :type language: str
    :ivar email: email of the User (defaults to '')
    :type email: str
    """

    def __init__(self, json, **kwargs):
        """Construct a user from provided json data."""
        super().__init__(json, **kwargs)

        self.username = self._json_data.get("username", "")
        self.id = self._json_data.get("pk", "")

    def __repr__(self):  # pragma: no cover
        return f"<pyke {self.__class__.__name__} '{self.username}' id {self.id}>"

    @property
    def default_name(self) -> str:
        """
        Get default name, prioritizing the user name over the KE-chain name.

        :return: Name
        :rtype str
        """
        return self.username if self.username else self.name

    @property
    def timezone(self) -> pytz.BaseTzInfo:
        """
        Timezone of the user.

        Defaults to timezone UTC.
        With return a pytz timezone eg. 'Europe/Amsterdam'

        :return: timezone object (compatible with datetime)
        :rtype: TzInfo
        """
        return pytz.timezone(zone=self._json_data.get("timezone", "UTC"))

    @property
    def language(self) -> str:
        """
        Language code of the user.

        Defaults to English ('en") when no language code is configured.

        :return: language code string
        :rtype: basestring
        """
        return self._json_data.get("language_code", LanguageCodes.ENGLISH)

    @property
    def email(self) -> str:
        """
        Email of the user.

        :return: email address, default is empty string.
        :rtype: basestring
        """
        return self._json_data.get("email", "")

    def now_in_my_timezone(self) -> datetime.datetime:
        """
        Get current time in the timezone of the User.

        Defaults to timezone GMT+1 (Europe/Amsterdam).

        :return: Current datetime
        :rtype datetime.datetime
        """
        timezone_definition = self._json_data["timezone"]
        if timezone_definition:
            timezone = pytz.timezone(timezone_definition)
        else:
            # if there is no timezone set then the Europe/Amsterdam timezone
            timezone = pytz.timezone("Europe/Amsterdam")

        # Default is utc timezone
        utc_time = datetime.datetime.now(tz=pytz.utc)

        # Convert to local timezone
        local_time = utc_time.astimezone(timezone)

        return local_time



from typing import Optional, Union, List, Dict

import requests

from pykechain.enums import TeamRoles, ScopeStatus
from pykechain.exceptions import APIError
from pykechain.models.user import User
from .base import Base
from .input_checks import check_text, check_type, check_enum, check_user
from ..utils import Empty, empty, clean_empty_values


class Team(Base):
    """A virtual object representing a KE-chain Team.

    :ivar name: team name
    :ivar id: uuid of the team
    """

    def __init__(self, json, **kwargs):
        """Construct a team from provided json data."""
        super().__init__(json, **kwargs)

        self.ref = json.get("ref")
        self.description = json.get("description")
        self.options = json.get("options")
        self.is_hidden = json.get("is_hidden")

    def _update(self, resource, update_dict=None, params=None, **kwargs):
        """Update the team in-place."""
        url = self._client._build_url(resource, **kwargs)
        response = self._client._request("PUT", url, json=update_dict, params=params)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError(f"Could not update Team {self}", response=response)

        self.refresh(json=response.json().get("results")[0])

    def edit(
        self,
        name: Optional[Union[str, Empty]] = empty,
        description: Optional[Union[str, Empty]] = empty,
        options: Optional[Union[Dict, Empty]] = empty,
        is_hidden: Optional[Union[bool, Empty]] = empty,
        **kwargs,
    ) -> None:
        """
        Edit the attributes of the Team.

        :param name: (o) name of the Team
        :type name: str
        :param description: (o) description of the Team
        :type description: str
        :param options: (o) options dictionary to set attributes such as `landingPage`.
        :type options: dict
        :param is_hidden: flag to hide the Team
        :type is_hidden: bool
        :return: None
        :raises IllegalArgumentError whenever inputs are not of the correct type
        """
        update_dict = {
            "id": self.id,
            "name": check_text(name, "name") or self.name,
            "description": check_text(description, "description"),
            "options": check_type(options, dict, "options"),
            "is_hidden": check_type(is_hidden, bool, "is_hidden"),
        }

        if kwargs:  # pragma: no cover
            update_dict.update(kwargs)

        update_dict = clean_empty_values(update_dict=update_dict)

        self._update(resource="team", team_id=self.id, update_dict=update_dict)

    def delete(self) -> None:
        """
        Delete the team.

        Members of the team remain intact.

        :return: None
        """
        url = self._client._build_url(resource="team", team_id=self.id)
        response = self._client._request("DELETE", url=url)

        if response.status_code != requests.codes.no_content:  # pragma: no cover
            raise APIError(f"Could not delete Team {self}", response=response)

    def members(self, role: Optional[Union[TeamRoles, str]] = None) -> List[Dict]:
        """Members of the team.

        You may provide the role in the team, to retrieve only the team member with that role. Normally there is a
        single owner, that has administration rights of the team. Normal team members do not have any rights to
        administer the team itself such as altering the team name, team image and team members. Administrators do
        have the right to administer the the team members.

        :param role: (optional) member belonging to a role :class:`pykechain.enums.TeamRoles` to return.
        :type role: basestring or None
        :raises IllegalArgumentError: when providing incorrect roles
        :return: list of dictionaries with members (pk, username, role, email)

        Example
        -------
        >>> my_team = client.team(name='My own team')
        >>> my_team.members()
        [{"pk":1, "username"="first user", "role"="OWNER", "email":"email@address.com"}, ...]

        """
        check_enum(role, TeamRoles, "role")

        member_list = list(self._json_data.get("members"))
        if role:
            return [teammember for teammember in member_list if teammember.get("role") == role]
        else:
            return member_list

    def add_members(
        self,
        users: Optional[List[Union[User, str]]] = None,
        role: Optional[Union[TeamRoles, str]] = TeamRoles.MEMBER,
    ) -> None:
        """Members to add to a team.

        :param users: list of members, either `User` objects or usernames
        :type users: List of `User` or List of pk
        :param role: (optional) role of the users to add (default `TeamRoles.MEMBER`)
        :type role: basestring
        :raises IllegalArgumentError: when providing incorrect user information

        Example
        -------
        >>> my_team = client.team(name='My own team')
        >>> other_user = client.users(name='That other person')
        >>> myself = client.users(name='myself')
        >>> my_team.add_members([myself], role=TeamRoles.MANAGER)
        >>> my_team.add_members([other_user], role=TeamRoles.MEMBER)

        """
        update_dict = {
            "role": check_enum(role, TeamRoles, "role"),
            "users": [check_user(user, User, "users") for user in users],
        }

        self._update("team_add_members", team_id=self.id, update_dict=update_dict)

    def remove_members(self, users: Optional[List[Union[User, str]]] = None) -> None:
        """
        Remove members from the team.

        :param users: list of members, either `User` objects or usernames
        :type users: List of `User` or List of pk
        :raises IllegalArgumentError: when providing incorrect user information


        Example
        -------
        >>> my_team = client.team(name='My own team')
        >>> other_user = client.users(name='That other person')
        >>> my_team.remove_members([other_user])

        """
        update_dict = {"users": [check_user(user, User, "users") for user in users]}

        self._update("team_remove_members", update_dict=update_dict, team_id=self.id)

    def scopes(self, status: Optional[ScopeStatus] = None, **kwargs) -> List["Scope"]:
        """Scopes associated to the team."""
        return self._client.scopes(team=self.id, status=status, **kwargs)



import os
import tempfile
import warnings
from collections import namedtuple
from typing import Optional, List, Any

from pykechain import Client
from pykechain.enums import PropertyType, Multiplicity, Category
from pykechain.exceptions import IllegalArgumentError, NotFoundError, MultipleFoundError
from pykechain.models import Part, AnyProperty, Property, PropertyValueFilter
from pykechain.utils import temp_chdir

# global variable
__mapping_dictionary = None
__edited_one_many: list = list()
__references = dict()
__attachments = list()

_InstanceCopy = namedtuple(
    "InstanceCopyAttributes",
    field_names=[
        "instance_original",
        "model_original",
        "target_parent_instance",
        "name",
    ],
)


# TODO Deprecate original utility functions in July 2021


def get_mapping_dictionary(clean=False) -> dict:
    """
    Get a temporary helper to map some keys to some values. Mainly used in the relocation of parts and models.

    :param clean: (optional) boolean flag to reset the mapping dictionary
    :type clean: bool
    :return: singleton dictionary (persistent in the script) for mapping use.
    """
    global __mapping_dictionary
    if not __mapping_dictionary or clean:
        __mapping_dictionary = dict()
    return __mapping_dictionary


def get_edited_one_many(clean=False) -> list:
    """
    Get a temporary helper to help relocating Parts with one to many relationships.

    Only used in the relocation of parts and models.

    :param clean: (optional) boolean flag to reset the list of Parts
    :type clean: bool
    :return: singleton list (persistent in the script) for tracking purposes
    """
    global __edited_one_many
    if not __edited_one_many or clean:
        __edited_one_many = list()
    return __edited_one_many


def get_references(clean=False) -> dict:
    """
    Get a temporary helper dictionary to store references (values) per property (keys).

    :param clean: (optional) boolean flag to reset the list of references
    :type clean: bool
    :return: singleton dictionary (persistent in the script) for tracking purposes
    """
    global __references
    if not __references or clean:
        __references = dict()
    return __references


def get_attachments(clean=False) -> list:
    """
    Get a temporary helper list to store attachment properties.

    :param clean: (optional) boolean flag to reset the list
    :type clean: bool
    :return: singleton list (persistent in the script) for tracking purposes
    """
    global __attachments
    if not __attachments or clean:
        __attachments = list()
    return __attachments


def get_illegal_targets(part: Part, include: set):
    """
    Retrieve the illegal parent parts where `Part` can be moved/copied.

    :param part: `Part` to be moved/copied.
    :type part: :class:`Part`
    :param include: `Set` object with id's to be avoided as target parent `Part`
    :type include: set
    :return: `List` object of illegal id's
    :rtype: list
    """
    list_of_illegal_targets = include or set()
    list_of_illegal_targets.update({c.id for c in part.all_children()})
    return list_of_illegal_targets


def map_property_instances(original_part: Part, new_part: Part) -> None:
    """
    Map the id of the original part with the `Part` object of the newly created one.

    Updated the singleton `mapping dictionary` with the new mapping table values.

    :param original_part: `Part` object to be copied/moved
    :type original_part: :class:`Part`
    :param new_part: `Part` object copied/moved
    :type new_part: :class:`Part`
    :return: None
    """
    # Map the original part with the new one
    mapping = get_mapping_dictionary()
    mapping[original_part.id] = new_part

    # Do the same for each Property of original part instance, using the 'model' id and the get_mapping_dictionary
    for prop_original in original_part.properties:
        mapping[prop_original.id] = [
            prop_new
            for prop_new in new_part.properties
            if mapping[prop_original.model_id].id == prop_new.model_id
        ][0]


def relocate_model(
    part: Part,
    target_parent: Part,
    name: Optional[str] = None,
    include_children: Optional[bool] = True,
) -> Part:
    """
    Move the `Part` model under a target parent `Part` model.

    .. versionadded:: 2.3

    :param part: `Part` object to be moved
    :type part: :class:`Part`
    :param target_parent: `Part` object under which the desired `Part` is moved
    :type target_parent: :class:`Part`
    :param name: how the moved top-level `Part` should be called
    :type name: basestring
    :param include_children: True to move also the descendants of `Part`. If False, the children will be lost.
    :type include_children: bool
    :return: moved :class: Part model.
    :raises IllegalArgumentError: if target_parent is descendant of part
    """
    warnings.warn(
        "`relocate_model` is no longer in use and will be deprecated in July 2021.",
        PendingDeprecationWarning,
    )

    if target_parent.id in get_illegal_targets(part, include={part.id}):
        raise IllegalArgumentError(
            "Cannot relocate part `{}` under target parent `{}`, because the target is part of "
            "its descendants".format(part.name, target_parent.name)
        )

    if include_children:
        part.populate_descendants()

    # First, if the user doesn't provide the name, then just use the default "Clone - ..." name
    if not name:
        name = f"CLONE - {part.name}"

    # Recursively create the part model copy
    moved_part_model = move_part_model(
        part=part,
        target_parent=target_parent,
        name=name,
        include_children=include_children,
    )

    # Try to update references to parts by updating the UUID via the mapping dictionary
    Property.set_bulk_update(True)
    mapping = get_mapping_dictionary()
    for prop_old, references_old in get_references().items():  # type: (AnyProperty, list)
        prop_new = mapping.get(prop_old.id)

        # try to map to a new ID, default to the existing reference ID
        references_new = [mapping.get(r, r) for r in references_old]
        prop_new.value = references_new
    Property.update_values(client=part._client)

    return moved_part_model


def move_part_model(
    part: Part,
    target_parent: Part,
    name: str,
    include_children: bool,
) -> Part:
    """
    Copy the `Part` model under a target parent `Part` model, recursively.

    .. versionadded:: 2.3

    :param part: `Part` object to be moved
    :type part: :class:`Part`
    :param target_parent: `Part` object under which the desired `Part` is moved
    :type target_parent: :class:`Part`
    :param name: how the moved top-level `Part` should be called
    :type name: basestring
    :param include_children: True to move also the descendants of `Part`. If False, the children will be lost.
    :type include_children: bool
    :return: moved :class: Part model.
    """
    warnings.warn(
        "`move_part_model` is no longer in use and will be deprecated in July 2021.",
        PendingDeprecationWarning,
    )
    return _copy_part_model(
        part=part,
        target_parent=target_parent,
        name=name,
        include_children=include_children,
    )


def _copy_part_model(
    part: Part,
    target_parent: Part,
    name: str,
    include_children: bool,
) -> Part:
    """
    Copy the `Part` model under a target parent `Part` model, including its descendants recursively.

    .. versionadded:: 2.3

    :param part: `Part` object to be copied
    :type part: :class:`Part`
    :param target_parent: `Part` object under which the desired `Part` is copied
    :type target_parent: :class:`Part`
    :param name: how the copied top-level `Part` should be called
    :type name: basestring
    :param include_children: True to also copy the descendants of `Part`. If False, the children will be lost.
    :type include_children: bool
    :return: moved :class: Part model.
    """
    property_fvalues = list()
    for prop in part.properties:  # type: AnyProperty
        property_values = dict(
            name=prop.name,
            description=prop.description,
            property_type=prop.type,
            value=_get_property_value(prop),
            unit=prop.unit,
            value_options=prop._options,
        )
        property_values = {k: v for k, v in property_values.items() if v is not None}
        property_fvalues.append(property_values)

    # Create the model and its properties in a single request
    moved_part_model = part._client.create_model_with_properties(
        name=name,
        parent=target_parent,
        multiplicity=part.multiplicity,
        properties_fvalues=property_fvalues,
    )

    # Map the current IDs with newly created objects
    get_mapping_dictionary().update({part.id: moved_part_model})
    for prop, moved_prop in zip(part.properties, moved_part_model.properties):
        get_mapping_dictionary()[prop.id] = moved_prop

    # The description cannot be added when creating a model, so edit the model after creation.
    if part.description:
        moved_part_model.edit(description=str(part.description))

    # Recursively, copy the sub-tree of the model
    if include_children:
        for sub_part in part.children():
            _copy_part_model(
                part=sub_part,
                target_parent=moved_part_model,
                name=sub_part.name,
                include_children=include_children,
            )

    return moved_part_model


def relocate_instance(
    part: Part,
    target_parent: Part,
    name: Optional[str] = None,
    include_children: Optional[bool] = True,
) -> Part:
    """
    Move the `Part` instance under a target parent `Part` instance.

    .. versionadded:: 2.3

    :param part: `Part` object to be moved
    :type part: :class:`Part`
    :param target_parent: `Part` object under which the desired `Part` is moved
    :type target_parent: :class:`Part`
    :param name: how the moved top-level `Part` should be called
    :type name: basestring
    :param include_children: True to move also the descendants of `Part`. If False, the children will be lost.
    :type include_children: bool
    :return: moved :class: `Part` instance
    """
    warnings.warn(
        "`relocate_instance` is no longer in use and will be deprecated in July 2021.",
        PendingDeprecationWarning,
    )

    # First, if the user doesn't provide the name, then just use the default "Clone - ..." name
    if not name:
        name = f"CLONE - {part.name}"

    # Initially the model of the part needs to be recreated under the model of the target_parent. Retrieve them.
    part_model = part.model()
    target_parent_model = target_parent.model()

    if include_children:
        part_model.populate_descendants()

    # To relocate an instance, its model has to go first
    relocate_model(
        part=part_model,
        target_parent=target_parent_model,
        name=part_model.name,
        include_children=include_children,
    )
    get_references(clean=True)

    # This function will move the part instance under the target_parent instance, and its children if required.
    moved_instance = move_part_instance(
        part_instance=part,
        target_parent=target_parent,
        part_model=part_model,
        name=name,
        include_children=include_children,
    )

    # Try to update references to parts by updating the UUID via the mapping dictionary
    Property.set_bulk_update(True)
    mapping = get_mapping_dictionary()
    for prop_old, references_old in get_references().items():  # type: (AnyProperty, list)
        prop_new = mapping.get(prop_old.id)

        # try to map to a new ID, default to the existing reference ID
        references_new = [mapping.get(r, r) for r in references_old]
        prop_new.value = references_new
    Property.update_values(client=part._client)

    return moved_instance


def move_part_instance(
    part_instance: Part,
    target_parent: Part,
    part_model: Part,
    name: Optional[str] = None,
    include_children: Optional[bool] = True,
) -> Part:
    """
    Copy the `Part` instance to target parent and updates the properties based on the original part instance.

    .. versionadded:: 2.3

    :param part_instance: `Part` object to be moved
    :type part_instance: :class:`Part`
    :param part_model: `Part` object representing the model of part_instance
    :type part_model: :class: `Part`
    :param target_parent: `Part` object under which the desired `Part` is moved
    :type target_parent: :class:`Part`
    :param name: how the moved top-level `Part` should be called
    :type name: basestring
    :param include_children: True to move also the descendants of `Part`. If False, the children will be lost.
    :type include_children: bool
    :return: moved :class: `Part` instance
    """
    warnings.warn(
        "`move_part_instance` is no longer in use and will be deprecated in July 2021.",
        PendingDeprecationWarning,
    )

    if not name:
        name = part_instance.name

    # Retrieve the model of the future part to be created
    moved_model = get_mapping_dictionary()[part_model.id]

    # Now act based on multiplicity
    if moved_model.multiplicity == Multiplicity.ONE:
        # If multiplicity is 'Exactly 1', that means the instance was automatically created with the model.
        moved_instance = moved_model.instances(parent_id=target_parent.id)[0]

    elif moved_model.multiplicity == Multiplicity.ONE_MANY:
        # If multiplicity is '1 or more', that means one instance has automatically been created with the model.
        # This first instance has to be used, but only once. Therefore, store the model in a global list after doing so.
        if moved_model.id not in get_edited_one_many():
            moved_instance = moved_model.instances(parent_id=target_parent.id)[0]
            get_edited_one_many().append(moved_model.id)
        else:
            moved_instance = target_parent.add(name=name, model=moved_model, suppress_kevents=True)
    else:
        # If multiplicity is '0 or more' or '0 or 1', it means no instance has been created automatically with the
        # model, so then everything must be created and then updated.
        moved_instance = target_parent.add(name=name, model=moved_model, suppress_kevents=True)

    # Update properties of the instance
    map_property_instances(original_part=part_instance, new_part=moved_instance)
    update_part_with_properties(part_instance, moved_instance, name=str(name))

    if include_children:
        sub_models = {child.id: child for child in part_model.children()}
        # Recursively call this function for every descendant. Keep the name of the original sub-instance.

        for sub_instance in part_instance.children():
            move_part_instance(
                part_instance=sub_instance,
                target_parent=moved_instance,
                part_model=sub_models.get(sub_instance.model_id, sub_instance.model()),
                name=sub_instance.name,
                include_children=True,
            )

    return moved_instance


def update_part_with_properties(
    part_instance: Part,
    moved_instance: Part,
    name: Optional[str] = None,
) -> Part:
    """
    Update the properties of the `moved_instance` based on the original `part_instance`.

    :param part_instance: `Part` object to be copied
    :type part_instance: :class:`Part`
    :param moved_instance: `Part` object copied
    :type moved_instance: :class:`Part`
    :param name: Name of the updated part
    :type name: basestring
    :return: moved :class: `Part` instance
    """
    warnings.warn(
        "`update_part_with_properties` is no longer in use and will be deprecated in July 2021.",
        PendingDeprecationWarning,
    )

    # Instantiate an empty dictionary used to collect all property values in order to update the part in one go.
    properties_id_dict = dict()
    for prop_instance in part_instance.properties:  # type: AnyProperty
        moved_prop_instance = get_mapping_dictionary()[prop_instance.id]

        # Do different magic if there is an attachment property and it has a value
        if prop_instance.type == PropertyType.ATTACHMENT_VALUE:
            if prop_instance.has_value():
                with temp_chdir() as target_dir:
                    full_path = os.path.join(target_dir or os.getcwd(), prop_instance.filename)
                    prop_instance.save_as(filename=full_path)
                    moved_prop_instance.upload(full_path)
            else:
                moved_prop_instance.clear()

        # For a reference value property, add the id's of the part referenced {property.id: [part1.id, part2.id, ...]},
        # if there is part referenced at all.
        elif prop_instance.type == PropertyType.REFERENCES_VALUE:
            if prop_instance.has_value():
                get_references()[prop_instance] = prop_instance.value_ids()

        elif prop_instance.type == PropertyType.SINGLE_SELECT_VALUE:
            if prop_instance.model().options:
                if prop_instance.value in prop_instance.model().options:
                    properties_id_dict[moved_prop_instance.id] = prop_instance.value

        elif prop_instance.type == PropertyType.MULTI_SELECT_VALUE:
            if prop_instance.has_value():
                if all(value in prop_instance.model().options for value in prop_instance.value):
                    properties_id_dict[moved_prop_instance.id] = prop_instance.value

        else:
            properties_id_dict[moved_prop_instance.id] = prop_instance.value

    # Update the name and property values in one go.
    moved_instance.update(
        name=str(name), update_dict=properties_id_dict, bulk=True, suppress_kevents=True
    )

    return moved_instance


def _copy_part(
    part: Part,
    target_parent: Part,
    name: Optional[str] = None,
    include_children: Optional[bool] = True,
    include_instances: Optional[bool] = True,
) -> Part:
    """
    Copy `part` below `target_parent`, optionally including all child Parts.

    Distinguish between `part` of different category:
    Part MODELS simply need a copy of themselves, and optionally copies of their instances.
    Part INSTANCES first need to have a copy of their part MODEL prior to a copy of themselves.

    :param part: Part to copy
    :param target_parent: Part to copy below
    :param name: (O) new name of the `part` copy
    :param include_children: (O) include the descendants of `part`, defaults to True
    :param include_instances: (O) In case of `part` being of category MODEL, include the instance Parts of that model.
        WARNING: By default, every instance is created per instance of the `target_parent`.
    :return: copy of `part`
    :rtype Part
    """
    get_mapping_dictionary(clean=True)
    get_edited_one_many(clean=True)
    get_references(clean=True)
    get_attachments(clean=True)

    if part.category == Category.INSTANCE:
        model = part.model()
        target_parent_model = target_parent.model()
        name_model = model.name

        instances = [
            _InstanceCopy(
                instance_original=part,
                target_parent_instance=target_parent,
                model_original=model,
                name=part.name if name is None else name,
            )
        ]

    else:  # part.category == Category.MODEL

        model = part
        target_parent_model = target_parent
        name_model = part.name if name is None else name

        if include_instances:
            try:
                target_parent_instance = target_parent_model.instance()
            except NotFoundError:
                raise IllegalArgumentError(
                    "Cannot copy part model `{}` including instances, since the target_parent"
                    " model `{}` has no instance to act as parent for the instances.".format(
                        model, target_parent_model
                    )
                )
            except MultipleFoundError:
                raise IllegalArgumentError(
                    "Cannot copy part model `{}` including instances, since the target_parent"
                    " model `{}` has multiple instances, making the parent for the instances"
                    " ambiguous.".format(model, target_parent_model)
                )

            instances = [
                _InstanceCopy(
                    instance_original=instance,
                    target_parent_instance=target_parent_instance,
                    model_original=model,
                    name=instance.name,
                )
                for instance in model.instances()
            ]
        else:
            instances = []

    # Verify if the target_parent is not below the part
    model.populate_descendants()
    if target_parent_model.id in get_illegal_targets(model, include={model.id}):
        raise IllegalArgumentError(
            "Cannot relocate part `{}` under target parent `{}`, because the target is part of "
            "its descendants".format(model.name, target_parent.name)
        )

    copied_model = _copy_part_model(
        part=model,
        target_parent=target_parent_model,
        name=name_model,
        include_children=include_children,
    )

    Property.set_bulk_update(True)
    _update_references()
    Property.update_values(client=copied_model._client, use_bulk_update=True)

    copied_instances = _copy_instances_recursive(
        client=copied_model._client,
        instances=instances,
        include_children=include_children,
    )

    mapping = get_mapping_dictionary()
    attachment_properties = get_attachments()
    for prop_original in attachment_properties:
        prop_new = mapping[prop_original.id]
        if prop_original.has_value():
            with tempfile.TemporaryDirectory() as target_dir:
                full_path = os.path.join(target_dir, prop_original.filename)
                prop_original.save_as(filename=full_path)
                prop_new.upload(full_path)

    _update_references()
    Property.update_values(client=copied_model._client)

    return copied_model if part.category == Category.MODEL else copied_instances[0]


def _copy_instances_recursive(
    client: Client,
    instances: List[_InstanceCopy],
    include_children: bool,
) -> List[Part]:
    """
    Create new Part instances in bulk, recursively.

    Reference and Attachment properties have to be updated outside this function.

    :param client: Client object
    :param instances: list of _Instance instances.
    :param include_children: whether to create instance parts
    :return: list of new Part instances
    :rtype list
    """
    if not instances:
        return []

    mapping = get_mapping_dictionary()
    create_request = []  # request for the bulk create
    created_instances_indices = []  # indices in a list
    original_instances = []  # part instances that require a copy
    new_instances = []  # all new Part objects

    for index, i in enumerate(instances):
        model_new = get_mapping_dictionary()[i.model_original.id]
        existing_instance = None

        if model_new.multiplicity == Multiplicity.ONE:
            # If multiplicity is 'Exactly 1', that means the instance was automatically created with the model.
            existing_instance = model_new.instances(parent_id=i.target_parent_instance.id)[0]

        elif model_new.multiplicity == Multiplicity.ONE_MANY:
            # If multiplicity is '1 or more', that means one instance has automatically been created with the model.
            # This first instance has to be used, but only once. Therefore, store the model in a global list after
            # doing so.
            if model_new.id not in get_edited_one_many():
                existing_instance = model_new.instances(parent_id=i.target_parent_instance.id)[0]
                get_edited_one_many().append(model_new.id)
        else:
            # If multiplicity is '0 or more' or '0 or 1', no instance has been created automatically with the model.
            pass

        if existing_instance:
            new_instances.append(existing_instance)
            map_property_instances(original_part=i.instance_original, new_part=existing_instance)

            if i.name != existing_instance.name:
                existing_instance.edit(name=i.name)

            # part already exists, but properties need to be updated
            for prop_original in i.instance_original.properties:
                prop = mapping.get(prop_original.id)
                prop.value = _get_property_value(prop_original)
        else:
            new_instances.append(None)
            created_instances_indices.append(index)
            original_instances.append(i)

            properties = []
            for prop in i.instance_original.properties:  # type: AnyProperty
                prop_value = _get_property_value(prop)
                if prop_value is not None:
                    properties.append(
                        dict(
                            name=prop.name,
                            value=prop_value,
                            model_id=mapping[prop.model_id].id,
                        )
                    )

            create_request.append(
                dict(
                    name=i.name,
                    parent_id=i.target_parent_instance.id,
                    model_id=model_new.id,
                    properties=properties,
                )
            )

    if create_request:
        created_instances = client._create_parts_bulk(
            parts=create_request,
            asynchronous=False,
            retrieve_instances=True,
        )
        for index, new_instance, i in zip(
            created_instances_indices, created_instances, original_instances
        ):  # type: int, Part, _InstanceCopy
            new_instances[index] = new_instance
            map_property_instances(original_part=i.instance_original, new_part=new_instance)

    if include_children:
        child_instances = []
        for i, new_instance in zip(instances, new_instances):
            child_models = {c.id: c for c in i.model_original.children()}

            for child_instance in i.instance_original.children():
                child_instances.append(
                    _InstanceCopy(
                        instance_original=child_instance,
                        model_original=child_models[child_instance.model_id],
                        target_parent_instance=new_instance,
                        name=child_instance.name,
                    )
                )

        _copy_instances_recursive(
            client=client,
            instances=child_instances,
            include_children=include_children,
        )

    return new_instances


def _update_references() -> None:
    """
    Set part reference values found when copying the Part(s).

    Part reference properties referring to child parts in the provided part tree should be updated to refer to
    the new child parts in the new tree. References to parts outside the provided tree can remain identical.
    Therefore, try to update the ID of the reference via the mapping dictionary.
    """
    mapping = get_mapping_dictionary()

    for (
        prop_original,
        references_original,
    ) in get_references().items():  # type: AnyProperty, List[Text]
        prop_new = mapping.get(prop_original.id)

        # Try to map to a new Part, default to the existing reference ID itself.
        references_new = [mapping.get(r, r) for r in references_original]
        if prop_new.type == PropertyType.REFERENCES_VALUE and references_new:

            if len(references_original) == 0:
                continue

            # Update the scope_id to match the new referenced part model
            # For parts copied in the same subtree
            if isinstance(references_new[0], Part):
                prop_new_options = {"scope_id": references_new[0].scope_id}
                prop_value = [ref.id for ref in references_new]
            # For parts left in the project from which they are copied from
            else:
                prop_value = list()
                for part_id in references_new:
                    referenced_part = prop_new._client.part(pk=part_id, category=prop_new.category)
                    prop_value.append(referenced_part.id)
                prop_new_options = {
                    "scope_id": prop_new._client.part(
                        pk=references_new[0], category=prop_new.category
                    ).scope_id
                }

            # Make sure the excluded property models ids are mapped on the new referenced part
            # model
            original_excluded_propmodel_ids = prop_original.get_excluded_propmodel_ids()
            if original_excluded_propmodel_ids:
                prop_new_options.update(
                    {
                        "propmodels_excl": [
                            mapping.get(r).id
                            for r in original_excluded_propmodel_ids
                            if r in mapping
                        ]
                    }
                )

            # Remap the prefilters on new referenced part model.
            original_prefilters = prop_original.get_prefilters()
            if original_prefilters:
                new_prefilters = [
                    PropertyValueFilter(
                        value=ovf.value,
                        filter_type=ovf.type,
                        property_model=mapping.get(ovf.id, ovf.id),
                    )
                    for ovf in original_prefilters
                ]
                prop_new_options.update(PropertyValueFilter.write_options(filters=new_prefilters))

            # Update the value and options (prefilters, excluded prop models and scope_id) in bulk
            if prop_new.category == Category.MODEL:
                prop_new.edit(value=prop_value, options=prop_new_options)
            else:
                prop_new.value = prop_value
        else:
            prop_new.value = references_new

    get_references(clean=True)


def _get_property_value(prop: AnyProperty) -> Any:
    """
    Get the property value, if directly applicable.

    In case of reference and attachment properties, the value has to be applied later via global attributes.

    :param prop: Any Property object
    :return: Any value
    """
    prop_value = None
    if prop.type in (
        PropertyType.REFERENCES_VALUE,
        PropertyType.ACTIVITY_REFERENCES_VALUE,
        PropertyType.SCOPE_REFERENCES_VALUE,
        PropertyType.TEAM_REFERENCES_VALUE,
        PropertyType.SERVICE_REFERENCES_VALUE,
    ):
        get_references()[prop] = prop.value_ids() if prop.has_value() else []
    elif prop.type == PropertyType.USER_REFERENCES_VALUE:
        get_references()[prop] = prop.value if prop.has_value() else []
    elif prop.type == PropertyType.ATTACHMENT_VALUE:
        get_attachments().append(prop)
    else:
        prop_value = prop._value

    return prop_value



from pykechain.exceptions import NotFoundError, IllegalArgumentError, APIError
from tests.classes import TestBetamax


class TestPartsReorderProperties(TestBetamax):
    # new in 1.10
    def test_reorder_properties_using_names(self):
        # Retrieve the front fork model
        front_fork_model = self.project.model("Front Fork")

        # Sort the properties of the front fork model based on their original order
        original_sorted_list_of_props = sorted(
            front_fork_model.properties, key=lambda k: k._json_data["order"]
        )

        # Make a new list with only the names of the properties sorted by order. This will later be used to
        # reverse the order to the initial status
        original_list_of_prop_names = [prop.name for prop in original_sorted_list_of_props]

        # Instantiate the list that will be used to reorder the properties
        desired_order_list = ["Material", "Height", "Color"]

        # Make the call to re-order the properties according to the above list
        front_fork_model.order_properties(property_list=desired_order_list)

        # Re-retrieve the front fork model
        front_fork_model = self.project.model("Front Fork")

        # Do the same steps as above. This will be used to check whether the action performed correctly.
        new_sorted_list_of_props = sorted(
            front_fork_model.properties, key=lambda k: k._json_data["order"]
        )
        new_list_of_prop_names = [prop.name for prop in new_sorted_list_of_props]

        # Check the new list with the desired one
        self.assertEqual(desired_order_list, new_list_of_prop_names)

        # Return the front fork model to the initial status
        front_fork_model.order_properties(property_list=original_list_of_prop_names)

    def test_reorder_properties_using_objects(self):
        # Retrieve the front fork model
        front_fork_model = self.project.model("Front Fork")

        # Sort the properties of the front fork model based on their original order
        original_sorted_list_of_props = sorted(
            front_fork_model.properties, key=lambda k: k._json_data["order"]
        )

        # Instantiate the list that will be used to reorder the properties
        desired_order_list = [
            front_fork_model.property("Material"),
            front_fork_model.property("Height"),
            front_fork_model.property("Color"),
        ]

        # Make the call to re-order the properties according to the above list
        front_fork_model.order_properties(property_list=desired_order_list)

        # Re-retrieve the front fork model
        front_fork_model = self.project.model("Front Fork")

        # Create a list of property id's, after the properties have been sorted.
        # This will be used to check whether the action performed correctly.
        new_sorted_list_of_props = sorted(
            front_fork_model.properties, key=lambda k: k._json_data["order"]
        )
        new_list_of_prop_ids = [prop.id for prop in new_sorted_list_of_props]

        # Create a list of property id's, based on the desired order.
        desired_order_list_ids = [prop.id for prop in desired_order_list]

        # Check the new list with the desired one
        self.assertEqual(desired_order_list_ids, new_list_of_prop_ids)

        # Return the front fork model to the initial status
        front_fork_model.order_properties(property_list=original_sorted_list_of_props)

    def test_reorder_wrong_properties(self):
        # Retrieve the front fork model
        front_fork_model = self.project.model("Front Fork")

        # Instantiate a wrong list that will be used to reorder the properties.
        desired_order_list = ["Material", "Height", "Color", "Width (mm)"]

        with self.assertRaises(NotFoundError):
            front_fork_model.order_properties(property_list=desired_order_list)

    def test_reorder_not_list(self):
        # Retrieve the front fork model
        front_fork_model = self.project.model("Front Fork")

        # Instantiate a wrong list that will be used to reorder the properties.
        desired_order_list = "Material Height Color"

        with self.assertRaises(IllegalArgumentError):
            front_fork_model.order_properties(property_list=desired_order_list)

    def test_reorder_properties_of_instance(self):
        # Retrieve the front fork model
        front_fork_instance = self.project.part(name="Front Fork", category="INSTANCE")

        # Instantiate a list that will be used to reorder the properties.
        desired_order_list = ["Material", "Height", "Color"]

        with self.assertRaises(APIError):
            front_fork_instance.order_properties(property_list=desired_order_list)
